#!/usr/local/bin/python3
from math import radians, sqrt, sin, cos, tan, pi
from random import random
from time import strftime

# TEST SELECTORS
# --------------------------------------------------------------------
# test g1        0 NIL > just > test_g2
# 1 test_rtm_g1  2 test_rtf_g1  3 test_urf_g1  4 test_exj_g1
# 5 test_exa_g1  6 test_xyj_g1  7 test_pta_g1  8 test_pra_g1
# 9 test_ISF_g1 10 test_Tmp_g1

# test g2        0 NIL > just > test_g3
# 1 test_rtm_g2  2 test_rtf_g2  3 test_urm_g2  4 test_exa_g2
# 5 test_xya_g2  6 test_xyj_g2  7 test_pta_g2  8 test_pra_g2
# 9 test_ISP_g2 10 test_ISF_g2

#      g3
# 1 rtm_repeat   2 urm_repeat   3 rtf_repeat   4 urf_repeat
# 5 exj_repeat   6 exa_repeat   7 pta_repeat   8 pra_repeat
# 9 ISF_repeat
# --------------------------------------------------------------------
debug = 0 # 1 is for debugging > removes exw adjustment that gives true outside dimensions
# SECTION: Inputs
# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
# TEST COMBINATIONS
g1               = 3
g2               = 4
g3               = 2

# BASICS
temperature,  standby_temp  = 245, 180
layer_height     = 0.24
extrusion_width  = 0.63        # ~~ nozzle dia
# FEATURES
twr1, twr2, twr3 = 1, 0, 1
grooves_no       = 0
grooved_sides_no = 1
repeat_ly        = 4           # layer repeat at constant settings, innermost "for" loop
zhop             = 0.12        # in start code

# for selection of "is_mode" in the reset section of the gcode file
ispi             = 3
# --------------------------------------------------------------------
# TEST VARIABLES
# INPUT SHAPING
is_modes = ['"ZVD"', '"ZVDD"', '"ZVDDD"', '"MZV"', '"EI2"', '"EI3"']
#              1        2         3          4        5        6
# ----------------
# All settings are reset to average of start & end values at start of output file.
# rpt include start & end values: increment = (end - start)/(rpt - 1)
# ....min 2 for all rpt
# ........if you use 3 everywere the middle layer of the final tests will be perfect

# Temperature
t_start,     t_end,   t_rpt = 245, 220, 5

# pressure advance | IS
pra_start, pra_end, pra_rpt = 0.000, 0.050, 3
isp                         = is_modes[ispi-1]
# input shaping frequency
isf_start, isf_end, isf_rpt = 36, 50, 8
isp_start, isp_end          =  1,  4

# retraction
# rtm
rtm_start, rtm_end, rtm_rpt = 0.80, 0.40, 3
# urm  ~ 0
urm_start, urm_end, urm_rpt = 0.00, 0.02, 3
# rtf
rtf_start, rtf_end, rtf_rpt = 3200, 4000, 3
# urf
urf_start, urf_end, urf_rpt = 2500, 3500, 3

# extruder dynamics
exj_start, exj_end, exj_rpt = 1500, 2500, 3
exa_start, exa_end, exa_rpt = 4000, 6000, 3

# xy dynamics
xyj_start, xyj_end, xyj_rpt =  600, 1400, 3
xya_start, xya_end, xya_rpt = 4000, 8000, 3
pta_start, pta_end, pta_rpt = 1500, 3500, 3
tra_start, tra_end, tra_rpt = 5000, 7000, 3

# --------------------------------------------------------------------
# PARAMETER BASICS
# speeds
# print speed for each tower in test layers
p_speed_1, p_speed_2, p_speed_3 = 40, 50, 70
# other speeds (mm/s)
travel_speed                = 175
layer_1_speed               =  25
layer_next_speed            =  35

fanspeed                    = 175  # test layers

# extrusion parameters (mm)
filament_dia                = 1.75
# --------------------------------------------------------------------
# walls & layers
walls_first                 = 4    # first layer | walls | perimeters
walls_next, layers_next     = 2, 3 # the next layers of the foundation
walls_test                  = 2    # walls on test layers

# --------------------------------------------------------------------
# SHAPES DIMENSIONS
# ie polygons: triangles to circles, with ability to ommit sides
# Bed Positioning,         note polygons print in sequence 1 2 3
# average centre of polygon tower 2, there is a random variation about this location
x_centre_p2                 = 0
y_centre_p2                 = 0
random_centres_scalar       = 1  # 1 is good

# location of centres of polygons 1 & 3 relative to polygon 2
xcentres                    =  45
ycentres                    = -20

# Polygons dimensioning (note "p_gaps_n")
p_side_1, p_side_2, p_side_3 = 25, 14, 25  # if == 0 not used, uses diameter

# Polygons diameter dimensioning
p_dia_1, p_sides_n_1, p_gaps_1, p_start_angle_1 = 25, 4, 0, 225
p_dia_2, p_sides_n_2, p_gaps_2, p_start_angle_2 = 20, 4, 0,  45
p_dia_3, p_sides_n_3, p_gaps_3, p_start_angle_3 = 25, 4, 0,  45

# GROOVES mm  no error checking for say too small spacing with many walls
g_end_1         = 3.0
g_end_1_taper   = 0.30  # fraction of extrusion width/layer
g_centre        = 10.0  # spacing between the centre of the 2 grooves on the outside wall
g_depth         = 2.5
g_width         = 2.0   # on outer layer
g_angle         = 90    # internal < 90  >> wide base
# --------------------------------------------------------------------
# LINE TWEAKING
overlap         = 0.04  # mm walls | perimeters overlap
side_close_adj  = 1.2   # *exw, shortens last line that closes polygons
std_line        = 1.00  # factor applied to normal walls
midline         = 1.4   # factor applied to outer wall width at change of another code
wideline        = 1.8   # factor applied to outer wall width at change of code range
primewidth      = 1.1
# --------------------------------------------------------------------
park_x_delta    = -20
park_y_delta    =  60
park_lift_delta =  20
# --------------------------------------------------------------------
# IMPORT DATA ... OVERIDES ALL RELEVANT ABOVE DATA | COMMENT OUT IF THIS FILE NOT USED
from d_settings  import * # overides above
# --------------------------------------------------------------------

# SECTION: INITIALISE DATA
# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
wide_speed = min(p_speed_1, p_speed_2, p_speed_3)/wideline # limits extreme flow vol
exw = extrusion_width
exw_ovlp = exw - overlap
end_2_last = 0  # .... ... EDIT check if still needed here ??

# BASE WALLS ADJUSTED TO GIVE ADEQUATE FOUNDATION
if grooves_no > 0:
   walls_additional = round(g_depth/exw) + 1    # to support grooves
else:
   walls_additional = 1

walls_next  = max(walls_next, walls_test + walls_additional)
walls_first = max(walls_first, walls_next)

# --------------------------------------------------------------------
# initial geometry
# randomise centre of polygon_2 from average centre (hollow elipse)
randradius = 2 + 3*random()
randangle  = 2*pi*random()
xoffset    = x_centre_p2 + (2*random_centres_scalar*randradius*sin(randangle))
yoffset    = y_centre_p2 + (1*random_centres_scalar*randradius*cos(randangle))
p_cnt_x_1  = xoffset - xcentres
p_cnt_y_1  = yoffset + ycentres
p_cnt_x_2  = xoffset
p_cnt_y_2  = yoffset
p_cnt_x_3  = xoffset + xcentres
p_cnt_y_3  = p_cnt_y_1

# radian conversion
p_start_angle_1 = radians(p_start_angle_1)
p_start_angle_2 = radians(p_start_angle_2)
p_start_angle_3 = radians(p_start_angle_3)

# g_angle    =  internal angle
g_angle_radn = radians(g_angle)
radian_90    = radians(90)
rad_180      = radians(180)

# ...tan_gvaDiv2 & tan_90_gva for angled grooves
# for groove sides
tan_gvaDiv2 = tan(g_angle_radn/2)
# for groove bases
tan_90_gva  = tan(radian_90 - g_angle_radn)
g_base      = g_width + exw + (2*g_depth*tan_90_gva)

# --------------------------------------------------------------------
# polygons calculated from p_side_n
# start with polygon diameter
p_rad_1 = p_dia_1/2
p_rad_2 = p_dia_2/2
p_rad_3 = p_dia_3/2

# IF p_side > 0         <<<  use p_side_n to derive p_rad(ius)
# override diameter input
if p_side_1 > 0:
    p_rad_1 = (p_side_1/2)/sin((2*pi/p_sides_n_1)/2)

if p_side_2 > 0:
    p_rad_2 = (p_side_2/2)/sin((2*pi/p_sides_n_2)/2)

if p_side_3 > 0:
    p_rad_3 = (p_side_3/2)/sin((2*pi/p_sides_n_3)/2)

# calc p_side_n which then is available for polygons
p_side_1 = (sin(pi/p_sides_n_1)*p_rad_1)*2
p_side_2 = (sin(pi/p_sides_n_2)*p_rad_2)*2
p_side_3 = (sin(pi/p_sides_n_3)*p_rad_3)*2
# --------------------------------------------------------------------
# initialise positons
curr_x     = 0.0
curr_y     = 0.0
curr_z     = layer_height
curr_layer = 1
# --------------------------------------------------------------------
# Mcode averages used when these parameters are not being tested
# set to their average in start code, hence useful to update these (depends??)
pra_avg = (pra_start + pra_end)/2
exj_avg = (exj_start + exj_end)/2
exa_avg = (exa_start + exa_end)/2
xyj_avg = (xyj_start + xyj_end)/2
xya_avg = (xya_start + xya_end)/2
pta_avg = (pta_start + pta_end)/2
tra_avg = (tra_start + tra_end)/2
rtm_avg = (rtm_start + rtm_end)/2
urm_avg = (urm_start + urm_end)/2
rtf_avg = (rtf_start + rtf_end)/2
urf_avg = (urf_start + urf_end)/2
# --------------------------------------------------------------------

# SECTION: Lines and Polygons
# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
# THE LINES
# basic functions
def extrusion_for_length(print_mm):
    area = ((exw - lyh)*lyh) + (lyh**2*pi/4)
    exm = (print_mm*area)/((filament_dia**2)*pi/4)
    return exm

def up():
    global curr_z, curr_layer
    curr_z += layer_height
    curr_layer += 1
    gcodef.write("G1 Z%.2f ; layer %i\n" % (curr_z, curr_layer))

# the foundation of this entire file
# writes a line of specified width -  lwidth > width/flow is proportional
#   to that for the specified height & width
#   ie provides overextrusion
def line_xye(x, y, lwidth):
    length = sqrt(x**2 + y**2)
    global curr_x, curr_y
    curr_x += x
    curr_y += y
    exm = extrusion_for_length(length)*lwidth
    gcodef.write("G1 X%.3f Y%.3f E%.5f\n" % (curr_x, curr_y, exm))
    global total_ex_mm
    total_ex_mm += exm

def line___f(speed):
    gcodef.write("G1 F%i\n" % (speed*60))

def goto(x, y):               # relative move
    global curr_x, curr_y
    curr_x += x
    curr_y += y
    gcodef.write("G1 X%.3f Y%.3f\n" % (x, y))

def goto_abs(x, y, speed):    # absolute move
    global curr_x, curr_y
    curr_x = x
    curr_y = y
    gcodef.write("G1 X%.3f Y%.3f F%i\n" % (x, y, speed*60))

# --------------------------------------------------------------------
# POLYGONS:  WRITE LINES
# side of polygons
# coordinates for clockwise side of polygon
def poly_side(_rad_, angle_n, angle_i, _width_):
    global delta_x, delta_y
    x_b = _rad_*sin(angle_n)           # beginning
    y_b = _rad_*cos(angle_n)
    x_e = _rad_*sin(angle_n + angle_i) # end
    y_e = _rad_*cos(angle_n + angle_i)
    delta_x = x_e - x_b
    delta_y = y_e - y_b
    line_xye(delta_x, delta_y, _width_)

# last side of polygon, has length shortened
#     reduces over-printing when finishing a polygon
#     called only for the last side that closes a polygon
# used in def polygon() etc
def poly_side_last(rad, angle_n, angle_i, _width_):
    global delta_x, delta_y
    x_b = rad*sin(angle_n)
    y_b = rad*cos(angle_n)
    x_e = rad*sin(angle_n + angle_i)
    y_e = rad*cos(angle_n + angle_i)
    angle_close = angle_n + angle_i/2 + radian_90
    x_close = exw*sin(angle_close)*side_close_adj  # decrement to remove bulge
    y_close = exw*cos(angle_close)*side_close_adj
    delta_x = (x_e - x_close) - x_b
    delta_y = (y_e - y_close) - y_b
    line_xye(delta_x, delta_y, _width_)

# single simple polygon with multiple walls (perimeters)
def polygon(cntX, cntY, _rad_p_, sides, sidegaps, startdegree, _speed_, travel_speed, walls_n, _width_):
    angle_s = startdegree
    angle_i = 2.0*pi/sides
    exw_rad = exw/cos(angle_i/2)
    exw_rad_ovlp = (exw - overlap)/cos(angle_i/2)
# so that inputs are for centre line of outside wall for debugging
    rad_inner = _rad_p_ - (1 - debug)*exw_rad/2 - exw_rad_ovlp*(walls_n - 1)
    for wall_i in range(walls_n):
        rad_wall_i = rad_inner + exw_rad_ovlp*wall_i
        goto_abs(cntX + (rad_wall_i*sin(angle_s)),
                 cntY + (rad_wall_i*cos(angle_s)), travel_speed)
        line___f(_speed_)
        if wall_i == 0:
            gcodef.write("G11\n")
            if walls_n > 1:
                gcodef.write(";TYPE:Support material\n")
        if wall_i == (walls_n - 1):
            gcodef.write(";TYPE:Solid infill\n")
        for sdi in range(sides - sidegaps):  # ..............writes one polygon
            angle_n = angle_s + angle_i*sdi
            if sdi == sides - 1:  # ...........................the closing side
                poly_side_last(rad_wall_i, angle_n, angle_i, _width_)
            else:  # ..................................all sides except closing
                poly_side(rad_wall_i, angle_n, angle_i, _width_)
    gcodef.write("G10\n")

# --------------------------------------------------------------------
# GROOVED POLYGONS
# grooved side segment <<--- poly start angle,  side included angle, segment direction
# -----------------------------
def line_gv_seg(angle_nn, angle_ii, angle_gseg, gdistance, _width_):
# angle and distance of specific line
    angle_gi = angle_nn + (angle_ii/2) + radian_90 + angle_gseg
    x_gv = gdistance*sin((angle_gi))
    y_gv = gdistance*cos((angle_gi))
    line_xye(x_gv, y_gv, _width_)

# -----------------------------
# grooved side [rad ~ radius]
def side_grooved(rad_sg, angle_n, angle_i, _end_1_, _g_centre_, _base_sg_, _g_depth_,
                 _width_, end_2_last):
    gside_length   = rad_sg*sin((angle_i/2))*2
    groove_width  = _base_sg_ - 2*(_g_depth_*tan_90_gva)   # ok mouth of groove
    groove_s      = _g_depth_/sin(g_angle_radn)              # angled groove side
    end_1       = _end_1_
    centre_length = _g_centre_ - groove_width
    gcodef.write(";    START of grooved side\n")
    line_gv_seg(angle_n, angle_i,         0,                     end_1, _width_)
    line_gv_seg(angle_n, angle_i, (rad_180 - g_angle_radn),   groove_s, _width_)
    line_gv_seg(angle_n, angle_i,      rad_180*0,          _base_sg_, _width_)
    line_gv_seg(angle_n, angle_i, - (rad_180 - g_angle_radn), groove_s, _width_)
#     gcodef.write("; ... ... ... ... end_2_last %i\n" % (end_2_last))
    if grooves_no == 1:
        end_2 = gside_length - end_1 - groove_width - (exw*side_close_adj*end_2_last) #LAST
        line_gv_seg(angle_n, angle_i,         0,           end_2, _width_)
    else:
        line_gv_seg(angle_n, angle_i,         0,             centre_length, _width_)
        line_gv_seg(angle_n, angle_i,   (rad_180 - g_angle_radn), groove_s, _width_)
        line_gv_seg(angle_n, angle_i,      rad_180*0,          _base_sg_, _width_)
        line_gv_seg(angle_n, angle_i, - (rad_180 - g_angle_radn), groove_s, _width_)
        end_2 = gside_length - end_1 - centre_length - (groove_width*2) - exw*side_close_adj*end_2_last
        line_gv_seg(angle_n, angle_i,         0,           end_2, _width_)
    gcodef.write(";    END  of grooved side\n")

# -----------------------------
# universal polygon with variable grooves now with poly write
def poly_universal(cntX, cntY, _rad_p_, sides, sidegaps, startangle, _speed_, travel_speed,
                walls_n, _width_, end_taper):
    angle_s      = startangle
    angle_i      = 2.0*pi/sides
    exw_tanD2    = exw*tan_gvaDiv2
    exw_tanD2_ovlp = (exw - overlap)/tan_gvaDiv2
    exw_rad      = exw/cos(angle_i/2)
    exw_rad_ovlp = (exw - overlap)/cos(angle_i/2)
# so that inputs are for centre line of outside wall for debugging
    rad_inner = _rad_p_ - (1 - debug)*exw_rad/2 - exw_rad_ovlp*(walls_n - 1)
# parameters for each segment of grooved sides of inside wall
    start_exw_ovlp  = exw_ovlp*tan(angle_i/2)
    end_exw_delta   = start_exw_ovlp + exw_tanD2_ovlp    # previous side + groove 1st side
    end_1_inner = g_end_1 - end_exw_delta*(walls_n - 1) + end_taper
    base_in      = g_base + ((exw_tanD2_ovlp)*2*(walls_n - 1))
    depth_in     = g_depth - (overlap*(walls_n - 1))     # inside walls depth reduced by overlap
# loop through walls
    for wall_i in range(walls_n):
        rad_wall_i = rad_inner + (exw_rad_ovlp*wall_i)
        goto_abs(cntX + (rad_wall_i*sin(angle_s)),
                 cntY + (rad_wall_i*cos(angle_s)), travel_speed)
        line___f(_speed_)
        end_1_i    = end_1_inner + end_exw_delta*wall_i
        depth_i    = depth_in + overlap*wall_i
        base_i     = base_in - (exw_tanD2_ovlp*wall_i*2)
        gcodef.write("; ... ... ... ... ... ... ... ... groove wall no %i\n" % (wall_i + 1))
        if wall_i == 0:
            gcodef.write("G11\n")
        if wall_i == 0:
            if walls_n > 1:
                gcodef.write(";TYPE:Perimeter\n")
        if wall_i == (walls_n - 1):
            if _width_ == std_line:
                if end_taper == 0:
                   gcodef.write(";TYPE:Solid infill\n")
                else:
                    gcodef.write(";TYPE:External perimeter\n")
        if wall_i == (walls_n - 1):
            if _width_ == wideline:
                gcodef.write(";TYPE:Overhang perimeter\n")
            else:
                if _width_ == midline:
                    gcodef.write(";TYPE:Top solid infill\n")
# loop around all sides
        for sdi in range(sides - sidegaps):  # ...........writes one polygon
            gcodef.write("; ...   ...   ...   ...   ...   side no %i\n" % (sdi+1))
            # sets angle of this side
            angle_n = angle_s + angle_i*sdi  # .............start of side sdi
# write all grooved sides in range, starting side 2
            if sdi in range(1, grooved_sides_no + 1):
#                 gcodef.write(";      --%i-of-%i-- max grooved sides\n" % (sdi, grooved_sides_no))
# .... shortens end_2 if last side of polygon
                if sdi == sides -1:
                    end_2_last = 1
                else:
                    end_2_last = 0
#                 gcodef.write("; ... ... ... ... end_2_last %i\n" % (end_2_last))
                side_grooved(rad_wall_i, angle_n, angle_i, end_1_i, g_centre, base_i,
                             depth_i, _width_, end_2_last)
            else:
                # straight sides
                    # last side
                if sdi == sides - 1:
#                     gcodef.write(";      last side\n")
                    poly_side_last(rad_wall_i, angle_n, angle_i, _width_) # LAST SIDE
                else:  # ..................all sides except grooved and closing
                    poly_side(rad_wall_i, angle_n, angle_i, _width_)  # STRAIGHT SIDES
    gcodef.write("G10\n")

# --------------------------------------------------------------------

# SECTION: LAYERS FROM POLYGONS
# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
# Base layers - with uniform (eg lower) speeds and defineable number of walls
# travel spped = 2 * base speed
def baselayers_pnnn(_base_speed_, _base_walls_):
    if twr1 == 1:
        polygon(p_cnt_x_1, p_cnt_y_1, p_rad_1, p_sides_n_1, p_gaps_10, p_start_angle_1,
         _base_speed_, 2*_base_speed_, _base_walls_, std_line)

    if twr2 == 1:
        polygon(p_cnt_x_2, p_cnt_y_2, p_rad_2, p_sides_n_2, p_gaps_20, p_start_angle_2,
         _base_speed_, 2*_base_speed_, _base_walls_, std_line)

    if twr3 == 1:
        polygon(p_cnt_x_3, p_cnt_y_3, p_rad_3, p_sides_n_3, p_gaps_30, p_start_angle_3,
         _base_speed_, 2*_base_speed_, _base_walls_, std_line)
    up()

# Test layers
def poly_nnn(_width_):
    gcodef.write(";                               line width factor %.2f\n" % (_width_))
    if twr1 == 1:
        polygon(p_cnt_x_1, p_cnt_y_1, p_rad_1, p_sides_n_1, p_gaps_1,
              p_start_angle_1, p_speed_1, travel_speed, walls_test, _width_)

    if twr2 == 1:
        polygon(p_cnt_x_2, p_cnt_y_2, p_rad_2, p_sides_n_2, p_gaps_2,
              p_start_angle_2, p_speed_2, travel_speed, walls_test, _width_)

    if twr3 == 1:
        polygon(p_cnt_x_3, p_cnt_y_3, p_rad_3, p_sides_n_3, p_gaps_3,
              p_start_angle_3, p_speed_3, travel_speed, walls_test, _width_)
    up()

def poly_nnn_universal(_width_,taper):
    gcodef.write(";                        groove line width factor %.2f\n" % (_width_))
    if twr1 == 1:
        poly_universal(p_cnt_x_1, p_cnt_y_1, p_rad_1, p_sides_n_1, p_gaps_1,
                     p_start_angle_1, p_speed_1, travel_speed, walls_test, _width_, taper)

    if twr2 == 1:
        poly_universal(p_cnt_x_2, p_cnt_y_2, p_rad_2, p_sides_n_2, p_gaps_2,
                     p_start_angle_2, p_speed_2, travel_speed, walls_test, _width_, taper)

    if twr3 == 1:
        poly_universal(p_cnt_x_3, p_cnt_y_3, p_rad_3, p_sides_n_3, p_gaps_3,
                     p_start_angle_3, p_speed_3, travel_speed, walls_test, _width_, taper)
    up()

# --------------------------------------------------------------------

# SECTION:  TEST FUNCTIONS CORE BUILDING BLOCKS
# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
# innermost "for" loop, writes layers at constant settings
# REPEATS LAYERS

def _repeat_layers(std_line):
    repeat_ly_half = round(repeat_ly /2)
#    end_t_1 = g_end_1
#    gcodef.write(";end t 1 %.2f\n" % (end_t_1))
    endtprlayer = g_end_1_taper*exw
    # creates < shape over repeated layers
    for i in range(repeat_ly_half + 1):
        end_1_ti = endtprlayer*i                      # increments end_1
#         gcodef.write(";                                 end_1_ti %.2f\n" % (end_1_ti))
        poly_nnn_universal(std_line, end_1_ti)
    for i in range(1, repeat_ly_half + 1):
        end_1_ti = endtprlayer*(repeat_ly_half - i)  # decrements end_1
#         gcodef.write(";                                 end_1_ti %.2f\n" % (end_1_ti))
        poly_nnn_universal(std_line, end_1_ti)

# --------------------------------------------------------------------
# MCODE LOOPING FUNCTIONS
# --------------------------------------------------------------------
# g3   Mcodes - for inner loops

def _for_rtm_repeat():
#     gcodef.write(";                                   _rtm_g3\n")
    for rtmi in range(rtm_rpt):
        rtm = (rtmi/(rtm_rpt - 1))*(rtm_end - rtm_start) + rtm_start
        gcodef.write("M117 \" . . . . . . . . . . . | g3 .  rtm . %.2f\"\n" % rtm)
        gcodef.write("M207 S%.2f\n" % (rtm))
        _repeat_layers(std_line)

def _for_urm_repeat():
#     gcodef.write(";                                   _urm_g3\n")
    for urmi in range(urm_rpt):
        urm = (urmi/(urm_rpt - 1))*(urm_end - urm_start) + urm_start
        gcodef.write("M117 \" . . . . . . . . . . . | g3 .  urm . %.2f\"\n" % urm)
        gcodef.write("M207 R%.2f\n" % (urm))
        _repeat_layers(std_line)

def _for_rtf_repeat():
#     gcodef.write(";                                   _rtf_g3\n")
    for rtfi in range(rtf_rpt):
        rtf = (rtfi/(rtf_rpt - 1))*(rtf_end - rtf_start) + rtf_start
        gcodef.write("M117 \" . . . . . . . . . . . | g3 .  rtf . %i\"\n" % rtf)
        gcodef.write("M207 F%i\n" % (rtf))
        _repeat_layers(std_line)

def _for_urf_repeat():
#     gcodef.write(";                                   _urf_g3\n")
    for urfi in range(urf_rpt):
        urf = (urfi/(urf_rpt - 1))*(urf_end - urf_start) + urf_start
        gcodef.write("M117 \" . . . . . . . . . . . | g3 .  urf . %i\"\n" % urf)
        gcodef.write("M207 T%i\n" % (urf))
        _repeat_layers(std_line)

def _for_exj_repeat():
#     gcodef.write(";                                   _exj_g3\n")
    for exji in range(exj_rpt):
        exj = (exji/(exj_rpt - 1))*(exj_end - exj_start) + exj_start
        gcodef.write("M117 \" . . . . . . . . . . . | g3 .  exj . %i\"\n" % exj)
        gcodef.write("M566 E%i\n" % (exj))
        _repeat_layers(std_line)

def _for_exa_repeat():
#     gcodef.write(";                                   _exa_g3\n")
    for exai in range(exa_rpt):
        exa = (exai/(exa_rpt - 1))*(exa_end - exa_start) + exa_start
        gcodef.write("M117 \" . . . . . . . . . . . | g3 .  exa . %i\"\n" % exa)
        gcodef.write("M201 E%i\n" % (exa))
        _repeat_layers(std_line)

def _for_pta_repeat():
#     gcodef.write(";                                   _pta_g3\n")
    for ptai in range(pta_rpt):
        pta = (ptai/(pta_rpt - 1))*(pta_end - pta_start) + pta_start
        gcodef.write("M117 \" . . . . . . . . . . . | g3 .  pta . %i\"\n" % pta)
        gcodef.write("M204 P%i T%i\n" % (pta, tra_avg))
        _repeat_layers(std_line)

def _for_pra_repeat():
#     gcodef.write(";                                   _pra_g3\n")
    for prai in range(pra_rpt):
        pra = (prai/(pra_rpt - 1))*(pra_end - pra_start) + pra_start
        gcodef.write("M117 \" . . . . . . . . . . . | g3 .  pra . %.3f\"\n" % pra)
        gcodef.write("M572 D0 S%.3f\n" % pra)
        _repeat_layers(std_line)

def _for_InputShaping_F_repeat():
#     gcodef.write(";                                   _isf_g3\n")
    for isfi in range(isf_rpt):
        isf = (isfi/(isf_rpt - 1))*(isf_end - isf_start) + isf_start
        gcodef.write("M117 \" . . . . . . . . . . . | g3 .  isf . %.1f\"\n" % isf)
        gcodef.write("M593 F%.1f\n" % isf)
        _repeat_layers(std_line)

# --------------------------------------------------------------------
# g2   Mcodes - for second loops

#  51  | rtm |
def test_rtm_g2():
#     gcodef.write("M117 \"test_rtm_g2\"\n")
#     gcodef.write(";                                    test_rtm_g2\n")
    for rtmi in range(rtm_rpt):
        rtm = (rtmi/(rtm_rpt - 1))*(rtm_end - rtm_start) + rtm_start
        gcodef.write("M117 \" . . . . . . . . . . . . . . | g2 . rtm %.2f\"\n" % rtm)
        gcodef.write("M207 S%.2f\n" % (rtm))
        gcodef.write("M207\n")
        _inner_g3()
        if rtmi < rtm_rpt - 1:
            poly_nnn_universal(midline, 0)

#  52  | rtf |
def test_rtf_g2():
#     gcodef.write("M117 \"test_rtf_g2\"\n")
#     gcodef.write(";                                  | test_rtf_g2\n")
    for rtfi in range(rtf_rpt):
        rtf = (rtfi/(rtf_rpt - 1))*(rtf_end - rtf_start) + rtf_start
        gcodef.write("M117 \" . . . . . . . . . . . . . . | g2 . rtf %i\"\n" % rtf)
        gcodef.write("M207 F%i\n" % (rtf))
        gcodef.write("M207\n")
        _inner_g3()
        if rtfi < rtf_rpt - 1:
            poly_nnn_universal(midline, 0)

#  53  | urm |
def test_urm_g2():
#     gcodef.write("M117 \"test_urm_g2\"\n")
#     gcodef.write(";                                  | test_urm_g2\n")
    for urmi in range(urm_rpt):
        urm = (urmi/(urm_rpt - 1))*(urm_end - urm_start) + urm_start
        gcodef.write("M117 \" . . . . . . . . . . . . . . | g2 . urm %.2f\"\n" % urm)
        gcodef.write("M207 R%.2f\n" % (urm))
        gcodef.write("M207\n")
        _inner_g3()
        if urmi < urm_rpt - 1:
            poly_nnn_universal(midline, 0)

#  54  | exa |
def test_exa_g2():
#     gcodef.write("M117 \"test_exa_g2\"\n")
#     gcodef.write(";                                  | test_exa_g2\n")
    for exai in range(exa_rpt):
        exa = (exai/(exa_rpt - 1))*(exa_end - exa_start) + exa_start
        gcodef.write("M117 \" . . . . . . . . . . . . . . | g2 . exa %i\"\n" % exa)
        gcodef.write("M201 E%i\n" % (exa))
        gcodef.write("M201\n")
        _inner_g3()
        if exai < exa_rpt - 1:
            poly_nnn_universal(midline, 0)

#  55  | xya |
def test_xya_g2():
#     gcodef.write("M117 \"test_xya_g2\"\n")
#     gcodef.write(";                                  | test_xya_g2\n")
    for xyai in range(xya_rpt):
        xya = (xyai/(xya_rpt - 1))*(xya_end - xya_start) + xya_start
        gcodef.write("M117 \" . . . . . . . . . . . . . . | g2 . xya %i\"\n" % xya)
        gcodef.write("M201 X%i Y%i\n" % (xya, xya))
        gcodef.write("M201\n")
        _inner_g3()
        if xyai < xya_rpt - 1:
            poly_nnn_universal(midline, 0)

#  56  | xyj |
def test_xyj_g2():
#     gcodef.write("M117 \"test_xyj_g2\"\n")
#     gcodef.write(";                                  | test_xyj_g2\n")
    for xyji in range(xyj_rpt):
        xyj = (xyji/(xyj_rpt - 1))*(xyj_end - xyj_start) + xyj_start
        gcodef.write("M117 \" . . . . . . . . . . . . . . | g2 . xyj %i\"\n" % xyj)
        gcodef.write("M566 X%i Y%i\n" % (xyj, xyj))
        gcodef.write("M566\n")
        _inner_g3()
        if xyji < xyj_rpt - 1:
            poly_nnn_universal(midline, 0)

#  57 | pta|
def test_pta_g2():
#     gcodef.write("M117 \"test_pta_g2\"\n")
#     gcodef.write(";                                  | test_pta_g2\n")
    for ptai in range(pta_rpt):
        pta = (ptai/(pta_rpt - 1))*(pta_end - pta_start) + pta_start
        gcodef.write("M117 \" . . . . . . . . . . . . . . | g2 . pta %i\"\n" % pta)
        gcodef.write("M204 P%i T%i\n" % (pta, tra_avg))
        gcodef.write("M204\n")
        _inner_g3()
        if ptai < pta_rpt - 1:
            poly_nnn_universal(midline, 0)

#  58  | Pressure Advance    | |
def test_pra_g2():
#     gcodef.write("M117 \"test_pra_g2\"\n")
#     gcodef.write(";                                  | test_pra_g2\n")
    for prai in range(pra_rpt):
        pra = (prai/(pra_rpt - 1))*(pra_end - pra_start) + pra_start
        gcodef.write("M117 \" . . . . . . . . . . . . . . | g2 . pra %.3f\"\n" % pra)
        gcodef.write("M572 D0 S%.3f\n" % pra)
        gcodef.write("M572\n")
        _inner_g3()
        if prai < pra_rpt - 1:
            poly_nnn_universal(midline, 0)

#  59  | InputShaping P |
def test_ISP_g2():
#     gcodef.write("M117 \"test_ISP_g2\"\n")
#     gcodef.write(";                                  | test_ISP_g2\n")
#     is_modes = ['"ZVD"', '"ZVDD"', '"ZVDDD"', '"MZV"', '"EI2"', '"EI3"']
    for ispi in range(isp_start -1, (isp_end)):
        isp = is_modes[ispi]
        gcodef.write("M117 \" . . . . . . . . . . . . . . | g2 . isp %s\"\n" % isp)
        gcodef.write("M593 P%s\n" % isp)
        gcodef.write("M593\n")
        _inner_g3()
        if ispi < isp_end - 1:
            poly_nnn_universal(midline, 0)

#  510 | InputShaping F |
def test_ISF_g2():
#     gcodef.write("M117 \"test_ISF_g2\"\n")
#     gcodef.write(";                                    test_ISF_g2\n")
    for isfi in range(isf_rpt):
        isf = (isfi/(isf_rpt - 1))*(isf_end - isf_start) + isf_start
        gcodef.write("M117 \" . . . . . . . . . . . . . . | g2 . isf %.1f\"\n" % isf)
        gcodef.write("M593 F%.1f\n" % isf)
        gcodef.write("M593\n")
        _inner_g3()
        if isfi < isf_rpt - 1:
            poly_nnn_universal(midline, 0)

# --------------------------------------------------------------------
# g1   Mcodes - for outer of 3 loops

# 70 | Pressure Advance |
def test___0_g1():
    gcodef.write(";                                  | test___0_g1 | | |\n\n")
    _test_g2()

def test_pra_g1():
#     gcodef.write(";                                  | test_pra_g1 | | |\n")
    for prai in range(pra_rpt):  # ............... OUTER
        pra = (prai/(pra_rpt - 1))*(pra_end - pra_start) + pra_start
        gcodef.write("M117 \" .  .  .  .  .  .  .  .  .  .  .  .  .  g1 . pra %.3f\"\n" % pra)
        gcodef.write("M572 D0 S%.3f\n" % pra)
        gcodef.write("M572\n")
        _test_g2()
        if prai < pra_rpt - 1:
            poly_nnn_universal(wideline, 0)

# 71  | rtm |
def test_rtm_g1():
#     gcodef.write(";                                  | test_rtm_g1 | | |\n")
    for rtmi in range(rtm_rpt):
        rtm = (rtmi/(rtm_rpt - 1))*(rtm_end - rtm_start) + rtm_start
        gcodef.write("M117 \" .  .  .  .  .  .  .  .  .  .  .  .  .  g1 . rtm %.2f\"\n" % rtm)
        gcodef.write("M207 S%.2f\n" % (rtm))
        gcodef.write("M207\n")
        _test_g2()
        if rtmi < rtm_rpt - 1:
            poly_nnn_universal(wideline, 0)

# 72  | rtf |
def test_rtf_g1():
#     gcodef.write(";                                  | test_rtf_g1 | | |\n")
    for rtfi in range(rtf_rpt):
        rtf = (rtfi/(rtf_rpt - 1))*(rtf_end - rtf_start) + rtf_start
        gcodef.write("M117 \" .  .  .  .  .  .  .  .  .  .  .  .  .  g1 . rtf %i\"\n" % rtf)
        gcodef.write("M207 F%i\n" % (rtf))
        gcodef.write("M207\n")
        _test_g2()
        if rtfi < rtf_rpt - 1:
            poly_nnn_universal(wideline, 0)

# 73 | urf |
def test_urf_g1():
#     gcodef.write(";                                  | test_urf_g1 | | |\n")
    for urfi in range(urf_rpt):
        urf = (urfi/(urf_rpt - 1))*(urf_end - urf_start) + urf_start
        gcodef.write("M117 \" .  .  .  .  .  .  .  .  .  .  .  .  .  g1 . urf %i\"\n" % urf)
        gcodef.write("M207 T%i\n" % (urf))
        gcodef.write("M207\n")
        _test_g2()
        if urfi < urf_rpt - 1:
            poly_nnn_universal(wideline, 0)
 
# 74 | exj |
def test_exj_g1():
#     gcodef.write(";                                  | test_exf_g1 | | |\n")
    for exji in range(exj_rpt):
        exj = (exji/(exj_rpt - 1))*(exj_end - exj_start) + exj_start
        gcodef.write("M117 \" .  .  .  .  .  .  .  .  .  .  .  .  .  g1 . exj %i\"\n" % exj)
        gcodef.write("M566 E%i\n" % (exj))
        gcodef.write("M566\n")
        _test_g2()
        if exji < exj_rpt - 1:
            poly_nnn_universal(wideline, 0)

# 75 | exa |
def test_exa_g1():
#     gcodef.write(";                                  | test_exa_g1 | | |\n")
    for exai in range(exa_rpt):
        exa = (exai/(exa_rpt - 1))*(exa_end - exa_start) + exa_start
        gcodef.write("M117 \" .  .  .  .  .  .  .  .  .  .  .  .  .  g1 . exa %i\"\n" % exa)
        gcodef.write("M201 E%i\n" % (exa))
        gcodef.write("M201\n")
        _test_g2()
        if exai < exa_rpt - 1:
            poly_nnn_universal(wideline, 0)

# 77 | pta |
def test_pta_g1():
#     gcodef.write(";                                  | test_pta_g1 | | |\n")
    for ptai in range(pta_rpt):
        pta = (ptai/(pta_rpt - 1))*(pta_end - pta_start) + pta_start
        gcodef.write("M117 \" .  .  .  .  .  .  .  .  .  .  .  .  .  g1 . pta %i\"\n" % pta)
        gcodef.write("M204 P%i\n" % pta)
#         gcodef.write("M204 P%i T%i\n" % (pta, tra_avg))
        gcodef.write("M204\n")
        _test_g2()
        if ptai < pta_rpt - 1:
            poly_nnn_universal(wideline, 0)

# 76 | xyj |
def test_xyj_g1():
#     gcodef.write(";                                  | test_xyj_g1 | | |\n")
    for xyji in range(xyj_rpt):
        xyj = (xyji/(xyj_rpt - 1))*(xyj_end - xyj_start) + xyj_start
        gcodef.write("M117 \" .  .  .  .  .  .  .  .  .  .  .  .  .  g1 . xyj %i\"\n" % xyj)
        gcodef.write("M566 X%i Y%i\n" % (xyj, xyj))
        gcodef.write("M566\n")
        _test_g2()
        if xyji < xyj_rpt - 1:
            poly_nnn_universal(wideline, 0)

#  79  | InputShaping F |
def test_ISF_g1():
#     gcodef.write(";                                  | test_ISF_g1 | | |\n")
    for isfi in range(isf_rpt):
        isf = (isfi/(isf_rpt - 1))*(isf_end - isf_start) + isf_start
        gcodef.write("M117 \" .  .  .  .  .  .  .  .  .  .  .  .  .  g1 . ISF F %.1f\"\n" % isf)
        gcodef.write("M593 F%.1f\n" % isf)
        gcodef.write("M593\n")
        _test_g2()
        if isfi < isf_rpt - 1:
            poly_nnn_universal(wideline, 0)

# 710 | Temperature |
def test_Tmp_g1():
#     gcodef.write(";                                  | test_Tmp_g1 | | |\n")
    for ti in range(t_rpt):
        tmp = (ti/(t_rpt - 1))*(t_end - t_start) + t_start
        gcodef.write("M117 \" .  .  .  .  .  .  .  .  .  .  .  .  .  g1 . Tmp %i\"\n" % tmp)
        gcodef.write("G10 S%i\n" % tmp)
        gcodef.write("G10\n")
        _test_g2()
        if ti < t_rpt - 1:
            poly_nnn_universal(wideline, 0)

# --------------------------------------------------------------------

# SECTION: TEST SELECTORS
# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
# Mcodes - inner g3 "for loops"

if g3 ==  1:
    _inner_g3 = _for_rtm_repeat
    pg3 = "rtm_"

if g3 ==  2:
    _inner_g3 = _for_urm_repeat
    pg3 = "urm_"

if g3 ==  3:
    _inner_g3 = _for_rtf_repeat
    pg3 = "rtf_"

if g3 ==  4:
    _inner_g3 = _for_urf_repeat
    pg3 = "urf_"

if g3 ==  5:
    _inner_g3 = _for_exj_repeat
    pg3 = "exj_"

if g3 ==  6:
    _inner_g3 = _for_exa_repeat
    pg3 = "exa_"

if g3 ==  7:
    _inner_g3 = _for_pta_repeat
    pg3 = "pta_"

if g3 ==  8:
    _inner_g3 = _for_pra_repeat
    pg3 = "pra_"

if g3 ==  9:
    _inner_g3 = _for_InputShaping_F_repeat
    pg3 = "isf_"

# --------------------------------------------------------------------
# test g2
if g2 ==  0:
    _test_g2 = _inner_g3
    pg2 = "____"

if g2 ==  1:
    _test_g2 = test_rtm_g2
    pg2 = "rtm_"

if g2 ==  2:
    _test_g2 = test_rtf_g2
    pg2 = "rtf_"

if g2 ==  3:
    _test_g2 = test_urm_g2
    pg2 = "urm_"

if g2 ==  4:
    _test_g2 = test_exa_g2
    pg2 = "exa_"

if g2 ==  5:
    _test_g2 = test_xya_g2
    pg2 = "xya_"

if g2 ==  6:
    _test_g2 = test_xyj_g2
    pg2 = "xyj_"

if g2 ==  7:
    _test_g2 = test_pta_g2
    pg2 = "pta_"

if g2 ==  8:
    _test_g2 = test_pra_g2
    pg2 = "pra_"

if g2 ==  9:
    _test_g2 = test_ISP_g2
    pg2 = "isp_"

if g2 ==  10:
    _test_g2 = test_ISF_g2
    pg2 = "isf_"

# --------------------------------------------------------------------
if g1 ==  0:
#    test = _test_g2
    test = test___0_g1
    pg1 = "__O_"

if g1 ==  1:
    test = test_rtm_g1
    pg1 = "rtm_"

if g1 ==  2:
    test = test_rtf_g1
    pg1 = "rtf_"

if g1 ==  3:
    test = test_urf_g1
    pg1 = "urf_"

if g1 ==  4:
    test = test_exj_g1
    pg1 = "exj_"

if g1 ==  5:
    test = test_exa_g1
    pg1 = "exa_"

if g1 ==  6:
    test = test_xyj_g1
    pg1 = "xyj_"

if g1 ==  7:
    test = test_pta_g1
    pg1 = "pta_"

if g1 ==  8:
    test = test_pra_g1
    pg1 = "pra_"

if g1 ==  9:
    test = test_ISF_g1
    pg1 = "ISF_"

if g1 ==  10:
    test = test_Tmp_g1
    pg1 = "Tmp_"

# --------------------------------------------------------------------

# SECTION: OUTPUT
# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
# End of file
park_x    = xoffset + park_x_delta
park_y    = yoffset + park_y_delta
park_lift =           park_lift_delta
# --------------------------------------------------------------------
# file name parameters: testnocombinaton + towers + testMcodes + time + groove + lh
path = "/Users/garisalexander_1/PythonOut/"
# test combination
g123str = str(g1) + str(g2) + str(g3) + "_"
# grooves
gn = str(int(grooves_no))
gs = str(int(grooved_sides_no))
# towers
t1 = str(int(twr1))
t2 = str(int(twr2))
t3 = str(int(twr3))

ttt = t1 + t2 + t3
lh = str(int(layer_height*100))
pras  = "-pra" + str(int(pra_avg*1000))
rtms  = "-rtm" + str(int(rtm_avg*100))
exjs  = "-exj" + str(int(exj_avg))
exas  = "-exa" + str(int(exa_avg))

# simplify some names for file naming
rtfa = rtf_avg
urfa = urf_avg
rtma = rtm_avg
urma = urm_avg
lyh = layer_height
# --------------------------------------------------------------------
# WRITE FILE
gcodefile = path + g123str + pg1 + pg2 + pg3 + ttt + "_"+ "g" + gn + gs + "_lh" + lh + strftime("_%H%M%S") + ".gcode"
gcodef = open(gcodefile, "w")
# start code
gcodef.write("T0\n")
gcodef.write("G10 S%i\n" % temperature) # XXXX
gcodef.write("M106 S0\n")  # base layers fan speed 0
gcodef.write("G21\n")
gcodef.write("G90\n")
gcodef.write("M83\n")
# reset some
gcodef.write("M220 S100\n")
gcodef.write("M221 S100\n")
gcodef.write("M201 X%i Y%i E%i\n" % (xya_avg, xya_avg, exa_avg))
gcodef.write("M204 P%i T%i\n" %     (pta_avg, tra_avg))
gcodef.write("M566 X%i Y%i E%i\n" % (xyj_avg, xyj_avg, exj_avg))
gcodef.write("M572 D0 S%.3f\n" %     pra_avg)
gcodef.write("M207 S%.2f R%.2f F%i T%i Z%.2f\n" % (rtma, urma, rtfa, urfa, zhop))
gcodef.write("M593 P%s\n" % isp)
# print to DWC console
gcodef.write("M220\n")
gcodef.write("M221\n")
gcodef.write("M201\n")
gcodef.write("M204\n")
gcodef.write("M207\n")
gcodef.write("M566\n")
gcodef.write("M572\n")
gcodef.write("M593\n")

# prime nozzle
gcodef.write("G1 Z25 F300\n")
prime_y = min(p_cnt_y_1 - p_rad_1, p_cnt_y_2 - p_rad_2, p_cnt_y_3 - p_rad_3)
# gcodef.write("; > >  goto ABS\n")
goto_abs(p_cnt_x_1 - p_rad_1 - 5, prime_y - 20, travel_speed/2)
gcodef.write("G4 S9\n")
# Z for first layer
gcodef.write("G1 Z%.2f F150 ;   .  .  .  .  .  .layer %i\n" % (curr_z, curr_layer))

total_ex_mm = 0.0
prime_width = primewidth
primed = - p_cnt_x_1 + p_rad_1 + p_cnt_x_3 + p_rad_3 + 10
gcodef.write(";TYPE:Skirt\n")
line___f(layer_1_speed*0.5)
line_xye(primed, 0, prime_width)
line___f(layer_1_speed)
line_xye(     0, 2, prime_width)
line_xye(- primed, 0, prime_width)
line_xye(     0, 2, prime_width)
line_xye(primed, 0, prime_width)
line_xye(     0, 2, prime_width)
line_xye(- primed/2, 0, 0.8)
gcodef.write("G10\n")

# base layers (no polygon side gaps here for better adhesion)
# First Layer
p_gaps_10 = 0
p_gaps_20 = 0
p_gaps_30 = 0
baselayers_pnnn(layer_1_speed, walls_first)

# Next Foundation Layers
for i in range(layers_next):
    baselayers_pnnn(layer_next_speed, walls_next)

# Test Layers
gcodef.write("M106 S%i\n" % fanspeed)
# reset extrusion length to count test layers only
total_ex_mm = 0.0
test()
# End code
park_z = curr_z + park_lift
gcodef.write("G1 X%.3f Y%.3f Z%.2f F%i\n" % (park_x, park_y, park_z, 1500))
gcodef.write("G10 R%i\n" % standby_temp)
gcodef.write("M106 S0\n\n")
# gcodef.write("#Extrusion Length .  .  .  %i mm\n" % total_ex_mm)
gcodef.write("M117 \"Extrusion Length of Tests .  .  %i mm\"\n" % total_ex_mm)
gcodef.close()
